/*
ArrayOfNetworksGUI
by Cory McKay

CLASS NAME:		ArrayOfNetworksGUI
CLASS TYPE:		Swing GUI
DESCRIPTION:	Provides an interface for training networks to transcribe output from the front end
				and performing transcriptions using trained networks.
LIMITATIONS:	Windows-based disk access conventinos were used (namely the use of \ as the directory
					seporator). The load_networks_action and save_networks_action sub-classes of this
					class may need to be modified under other platforms.
				Only one hidden layer is permitted.
*/

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.File;

class ArrayOfNetworksGUI
	extends JFrame
{
	/* FIELDS ******************************************************************/


	// Array of networks and configuration data about the loaded array
	NNArray network_array;
	String loaded_network_type;
	String loaded_input_type;
	int loaded_number_instruments;
	int loaded_number_notes;
	int loaded_number_bins_partials;
	int loaded_min_note;
	int loaded_max_note;

	// Network I/O data
	double[][][] input_data;			// data sent into network input nodes
	double[][][] output_data;			// model output data
	Object additional_info;				// any additional useful data derived from parsing input_data
	double[][][] stored_results;		// the results of passing test data through the network

	// Dialog boxes
	JFileChooser network_loader;		// used to open network files
	JFileChooser data_loader;			// used to open data files
	JFileChooser saver;					// used to save networks
	JFileChooser MIDI_saver;			// used to save transcriptions
	NewNetworkArrayDialog new_dialog;	// used to define settings for new networks
	
	// GUI containers
	Container contentPane;
	JPanel network_panel;
	JPanel processing_panel;
	JPanel network_panelN;
	JPanel processing_panelN;
	JPanel network_panelS;
	JPanel processing_panelS;
	JScrollPane scrollpane_left;
	JScrollPane scrollpane_centre;
	JScrollPane scrollpane_right;
	JTextArea text_area_left;
	JTextArea text_area_centre;
	JTextArea text_area_right;

	// GUI interactive elements related to creating networks
	JButton create_network_button;
	JComboBox network_type_box;
	JComboBox input_type_box;
	JTextField number_bins_partials_field;
	JTextField number_instruments_field;
	JTextField lowest_pitch_field;
	JTextField highest_pitch_field;
	JButton load_network_button;
	JButton save_network_button;
	JButton display_network_button;

	// GUI interactive elements related to processing data through networks
	JButton load_input_data_button;
	JButton display_input_data_button;
	JButton display_model_output_button;
	JTextField number_iterations_field;
	JButton train_network_button;
	JButton load_model_output_button;
	JButton run_samples_button;
	JCheckBox randomize_training_checkbox;
	JCheckBox clean_output_checkbox;
	JCheckBox instrument_monophony_checkbox;
	JButton save_midi_file_button;


	/* CONSTRUCTOR *************************************************************/


	ArrayOfNetworksGUI(String title)
	{
		// Initizlize network and data to empty
		network_array = null;
		loaded_network_type = new String();
		loaded_input_type = new String();
		input_data = null;
		output_data = null;
		loaded_number_bins_partials = 0;
		loaded_number_instruments = 0;
		loaded_number_notes = 0;
		loaded_min_note = 0;
		loaded_max_note = 0;
		stored_results = null;

		// Configure frame and its dialog boxes
		setTitle(title);
		new_dialog = new NewNetworkArrayDialog(ArrayOfNetworksGUI.this);
		network_loader = new JFileChooser();
		network_loader.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		network_loader.setCurrentDirectory(new File("."));
		saver = new JFileChooser();
		saver.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		saver.setCurrentDirectory(new File("."));
		data_loader = new JFileChooser();
		data_loader.setCurrentDirectory(new File("."));
		MIDI_saver = new JFileChooser();
		MIDI_saver.setCurrentDirectory(new File("."));

		// Instantiate containers
		contentPane = getContentPane();
		network_panel = new JPanel(new BorderLayout());
		processing_panel = new JPanel(new BorderLayout());
		network_panelN = new JPanel();
		processing_panelN = new JPanel();
		network_panelS = new JPanel();
		processing_panelS = new JPanel();
		text_area_left = new JTextArea();
		text_area_centre = new JTextArea();
		text_area_right = new JTextArea();

		// Instantiate interactive elements of GUI related to creating networks
		create_network_button = new JButton("New Network Array");
		network_type_box = new JComboBox();
		input_type_box = new JComboBox();
		number_bins_partials_field = new JTextField("6", 2);
		number_instruments_field = new JTextField("1", 2);
		lowest_pitch_field = new JTextField("60", 3);
		highest_pitch_field = new JTextField("72", 3);
		load_network_button = new JButton("Load Network Array");
		save_network_button = new JButton("Save Network Array");
		display_network_button = new JButton("Display Network Array");

		// Load items into combo boxes
		network_type_box.addItem("a net/note/inst");
		network_type_box.addItem("a net/inst");
		network_type_box.addItem("a combined net");
		network_type_box.addItem("a limited net");
		network_type_box.addItem("instrument ID");
		input_type_box.addItem("f sorted by A");
		input_type_box.addItem("f sorted by f");
		input_type_box.addItem("f and A");
		input_type_box.addItem("filterbanks");

		// Instantiate interactive elements of GUI related to processing data through networks
		load_input_data_button = new JButton("Load Input Data");
		display_input_data_button = new JButton("Display Input Data");
		load_model_output_button = new JButton("Load Model Output");
		display_model_output_button = new JButton("Display Model Output");
		number_iterations_field = new JTextField("1000", 6);
		train_network_button = new JButton("Train Networks");
		run_samples_button = new JButton("Process Data");
		randomize_training_checkbox = new JCheckBox("", true);
		clean_output_checkbox = new JCheckBox("", true);
		instrument_monophony_checkbox = new JCheckBox("", true);
		save_midi_file_button = new JButton("Save Transcription");

		// Prevent inappropriate user actions
		create_network_button.setEnabled(true);
		load_network_button.setEnabled(true);
		save_network_button.setEnabled(false);
		display_network_button.setEnabled(false);
		load_input_data_button.setEnabled(false);
		display_input_data_button.setEnabled(false);
		display_model_output_button.setEnabled(false);
		train_network_button.setEnabled(false);
		load_model_output_button.setEnabled(false);
		run_samples_button.setEnabled(false);
		save_midi_file_button.setEnabled(false);

		// Cause buttons to respond to being pressed
		create_network_button.addActionListener(new create_networks_action());
		load_network_button.addActionListener(new load_networks_action());
		save_network_button.addActionListener(new save_networks_action());
		display_network_button.addActionListener(new display_networks_action());
		load_input_data_button.addActionListener(new load_input_data_action());
		display_model_output_button.addActionListener(new display_model_output_action());
		display_input_data_button.addActionListener(new display_input_action());
		load_model_output_button.addActionListener(new load_model_output_action());
		train_network_button.addActionListener(new train_network_action());
		run_samples_button.addActionListener(new run_samples_action());
		save_midi_file_button.addActionListener(new save_midi_action());

		// Configure the left text display area
		text_area_left.setLineWrap(false);
		text_area_left.setRows(35);
		text_area_left.setColumns(25);
		text_area_left.setEditable(false);
		text_area_left.setText("");
		scrollpane_left = new JScrollPane(text_area_left);

		// Configure the centre text display area
		text_area_centre.setLineWrap(false);
		text_area_centre.setRows(35);
		text_area_centre.setColumns(22);
		text_area_centre.setEditable(false);
		text_area_centre.setText("");
		scrollpane_centre = new JScrollPane(text_area_centre);

		// Configure the right text display area
		text_area_right.setLineWrap(false);
		text_area_right.setRows(35);
		text_area_right.setColumns(22);
		text_area_right.setEditable(false);
		text_area_right.setText("");
		scrollpane_right = new JScrollPane(text_area_right);

		// Add GUI components related to creating networks
		network_panelN.add(new JLabel("Net Type:"));
		network_panelN.add(network_type_box);
		network_panelN.add(new JLabel("Input Type:"));
		network_panelN.add(input_type_box);
		network_panelN.add(new JLabel("Bins / Partials:"));
		network_panelN.add(number_bins_partials_field);
		network_panelN.add(new JLabel("Instruments:"));
		network_panelN.add(number_instruments_field);
		network_panelN.add(new JLabel("Min Pitch:"));
		network_panelN.add(lowest_pitch_field);
		network_panelN.add(new JLabel("Max Pitch:"));
		network_panelN.add(highest_pitch_field);
		network_panelS.add(create_network_button);
		network_panelS.add(load_network_button);
		network_panelS.add(save_network_button);
		network_panelS.add(display_network_button);

		// Add GUI components related to processing data in the network
		processing_panelN.add(load_input_data_button);
		processing_panelN.add(load_model_output_button);
		processing_panelN.add(display_input_data_button);
		processing_panelN.add(display_model_output_button);
		processing_panelN.add(save_midi_file_button);
		processing_panelS.add(train_network_button);
		processing_panelS.add(new JLabel("Epochs:"));
		processing_panelS.add(number_iterations_field);
		processing_panelS.add(new JLabel("Randomize:"));
		processing_panelS.add(randomize_training_checkbox);
		processing_panelS.add(run_samples_button);
		processing_panelS.add(new JLabel("Clean Output:"));
		processing_panelS.add(clean_output_checkbox);
		processing_panelS.add(new JLabel("Monophony:"));
		processing_panelS.add(instrument_monophony_checkbox);

		// Cause all GUI components to be displayed
		processing_panel.add(processing_panelN, BorderLayout.NORTH);
		processing_panel.add(processing_panelS, BorderLayout.SOUTH);
		network_panel.add(network_panelN, BorderLayout.NORTH);
		network_panel.add(network_panelS, BorderLayout.SOUTH);
		contentPane.add(network_panel, BorderLayout.NORTH);
		contentPane.add(scrollpane_left, BorderLayout.WEST);
		contentPane.add(scrollpane_centre, BorderLayout.CENTER);
		contentPane.add(scrollpane_right, BorderLayout.EAST);
		contentPane.add(processing_panel, BorderLayout.SOUTH);

		// Cause program to quit when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				System.exit(0);
			}
		});
	}


	/* SUB-CLASSES *************************************************************/


	class create_networks_action extends AbstractAction
	// Display a dialog box allowing the user to set the configurations of a new array
	// of neural networks. If the user presses OK, a new NNArray is created, replacing
	// any existing one in memory. Its weights are randomly generated. No change occurs
	// if the cancel butotn is pressed.
	{
		public void actionPerformed(ActionEvent event)
		{
			// Collect useful data from GUI components
			int number_notes = 1 + Integer.parseInt(highest_pitch_field.getText()) - Integer.parseInt(lowest_pitch_field.getText());
			int number_instruments = Integer.parseInt(number_instruments_field.getText());
			int number_bins_partials = Integer.parseInt(number_bins_partials_field.getText());
			String network_type = new String((String) network_type_box.getSelectedItem());
			String input_type = new String((String) input_type_box.getSelectedItem());

			// Calculate number of networks and number output units needed
			int number_networks = 0;
			int number_output_nodes = 0;
			if (network_type.equals("a net/note/inst"))
			{
				number_networks = number_instruments * number_notes;
				number_output_nodes = 1;
			}
			else if (network_type.equals("a net/inst"))
			{
				number_networks = number_instruments;
				number_output_nodes = number_notes;
			}
			else if (network_type.equals("a combined net"))
			{
				number_networks = 1;
				number_output_nodes = number_instruments * number_notes;
			}
			else if (network_type.equals("a limited net"))
			{
				number_networks = 1;
				number_output_nodes = number_notes;
			}
			else if (network_type.equals("instrument ID"))
			{
				number_networks = 1;
				number_output_nodes = number_instruments;
			}

			// Calculate number of input nodes needed
			int number_input_nodes = 0;
			if (input_type.equals("f and A"))
				number_input_nodes = 2 * number_bins_partials;
			else if (input_type.equals("f sorted by A") || input_type.equals("f sorted by f") || input_type.equals("filterbanks"))
				number_input_nodes = number_bins_partials;

			// Set correct number of networks, input nodes and output nodes
			new_dialog.num_networks_field.setText(String.valueOf(number_networks));
			new_dialog.num_inputs_field.setText(String.valueOf(number_input_nodes));
			new_dialog.num_outputs_field.setText(String.valueOf(number_output_nodes));

			// Display dialog box
			new_dialog.show();
			
			// Take action if OK button pressed
			if (new_dialog.was_updated() == true)
			{
				// Create network
				network_array = new_dialog.generate_new_network_array();

				// Store configuration of currently loaded network
				loaded_network_type = network_type;
				loaded_input_type = input_type;
				loaded_number_bins_partials = number_bins_partials;
				loaded_number_instruments = number_instruments;
				loaded_number_notes = number_notes;
				loaded_min_note = Integer.parseInt(lowest_pitch_field.getText());
				loaded_max_note = Integer.parseInt(highest_pitch_field.getText());

				// Update what user is permitted to do
				save_network_button.setEnabled(true);
				display_network_button.setEnabled(true);
				load_input_data_button.setEnabled(true);
				display_input_data_button.setEnabled(false);
				train_network_button.setEnabled(false);
				load_model_output_button.setEnabled(false);
				display_model_output_button.setEnabled(false);
				run_samples_button.setEnabled(false);
				save_midi_file_button.setEnabled(false);

				// Display action succesful message
				text_area_left.setText("Array of " + network_array.length + " networks " +
					"succesfully created");
				text_area_centre.setText("");
				text_area_right.setText("");
			}
		}
	}


	class load_networks_action extends AbstractAction
	// Bring up a JFileChooser dialog box to load an array of pre-trained networks.
	{
		public void actionPerformed(ActionEvent event)
		{
			int loader_return = network_loader.showOpenDialog(ArrayOfNetworksGUI.this);
			if (loader_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				File selection = network_loader.getSelectedFile();
				NetworksFileReader reader = new NetworksFileReader();
				try
				{
					// Set all dialog boxes to this directory
					data_loader.setCurrentDirectory(network_loader.getCurrentDirectory());
					saver.setCurrentDirectory(network_loader.getCurrentDirectory());
					MIDI_saver.setCurrentDirectory(network_loader.getCurrentDirectory());

					// Read in network settings and change appropriate GUI fields and stored configurations
					Object[] general_data = reader.get_overall_network_data(selection.getPath() + "\\guide.dat");
					loaded_network_type = (String) general_data[0];
					loaded_input_type = (String) general_data[1];
					int[] overall_settings = (int[]) general_data[2];
					network_type_box.setSelectedItem(loaded_network_type);
					input_type_box.setSelectedItem(loaded_input_type);
					number_bins_partials_field.setText(String.valueOf(overall_settings[0]));
					int number_networks = overall_settings[1];
					number_instruments_field.setText(String.valueOf(overall_settings[2]));
					lowest_pitch_field.setText(String.valueOf(overall_settings[3]));
					highest_pitch_field.setText(String.valueOf(overall_settings[4]));
					loaded_number_bins_partials = Integer.parseInt(number_bins_partials_field.getText());
					loaded_number_instruments = Integer.parseInt(number_instruments_field.getText());
					loaded_number_notes = 1 + Integer.parseInt(highest_pitch_field.getText()) - Integer.parseInt(lowest_pitch_field.getText());
					loaded_min_note = Integer.parseInt(lowest_pitch_field.getText());
					loaded_max_note = Integer.parseInt(highest_pitch_field.getText());

					// Prepare data stuctures to hold loaded data
					NetworkSettings[] network_info = new NetworkSettings[number_networks];
					double[][] weights = new double[number_networks][];

					// Process network files
					for (int i = 0; i < number_networks; i++)
					{
						// Load data from file into network_data
						reader = new NetworksFileReader();
						reader.read_network_file(selection.getPath() + "\\nn" + i + ".dat");
						double[] network_data = reader.get_network_data(); // the contents of the network file
						
						// Load settings and weights
						network_info[i] = new NetworkSettings(network_data[0], network_data[1],
							(int) network_data[2], (int) network_data[3], (int) network_data[4], 0, 0);
						weights[i] = new double[network_data.length - 5];
						for (int j = 0; j < weights[i].length; j++)
							weights[i][j] = network_data[ j + 5 ];
					}

					// Set up array
					network_array = new NNArray(network_info, weights);

					// Update what user is permitted to do
					save_network_button.setEnabled(true);
					display_network_button.setEnabled(true);
					load_input_data_button.setEnabled(true);
					display_input_data_button.setEnabled(false);
					train_network_button.setEnabled(false);
					load_model_output_button.setEnabled(false);
					display_model_output_button.setEnabled(false);
					run_samples_button.setEnabled(false);
					save_midi_file_button.setEnabled(false);

					// Report success
					text_area_left.setText("Array of " + network_array.length + " networks " +
						"succesfully created");
					text_area_centre.setText("");
					text_area_right.setText("");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Inappropriate directory specified.\n");
					text_area_centre.setText("");
					text_area_right.setText("");
				}				
			}			
		}
	}

	class save_networks_action extends AbstractAction
	// Brings up a JFileChooser dialog box to save the currently loaded array of networks
	// to disk. Only directories may be selected, as a file will be created for each
	// network in the array. If a directory does not already exist, it will be created.
	// A file named guide.dat is created containing the number of networks saved, the number
	// of instruments, the minimum pitch and the maximum pitch. The networks are saved in files
	// named nni.dat, where i is a number from 0 to number of networks - 1.
	{
		public void actionPerformed(ActionEvent event)
		{
			int saver_return = saver.showSaveDialog(ArrayOfNetworksGUI.this);
			if (saver_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				// Set all dialog boxes to this directory
				network_loader.setCurrentDirectory(saver.getCurrentDirectory());
				data_loader.setCurrentDirectory(saver.getCurrentDirectory());
				MIDI_saver.setCurrentDirectory(saver.getCurrentDirectory());

				File selection = saver.getSelectedFile();
				FWriter writer = new FWriter();
				try
				{
					// Create directory if does not exist already
					if (selection.isFile() == false && selection.isDirectory() == false)
						selection.mkdir();

					// Write guide file
					String guide_data = new String(loaded_network_type + "\n" + loaded_input_type + "\n" +
						number_bins_partials_field.getText() + "\n" + String.valueOf(network_array.length) +
						"\n" + number_instruments_field.getText() +	"\n" + lowest_pitch_field.getText() +
						"\n" + highest_pitch_field.getText() + "\n");
					writer.write_file(new String(selection.getPath() + "\\guide.dat"), guide_data);
					
					// Write each network file
					for (int i = 0; i < network_array.length; i++)
					{
						writer.write_file(new String(selection.getPath() + "\\nn" + i + ".dat"),
										  network_array.get_controller(i).get_network_data());
					}

					// Display success message
					text_area_left.setText("Networks saved succesfully in:\n" +
					                       selection.getPath() + "\n");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Unable to save networks\n");
				}				
			}	
		}
	}


	class display_networks_action extends AbstractAction
	// Display the structure of the network in the left display area
	{
		public void actionPerformed(ActionEvent event)
		{
			String[] network_contents = network_array.expose_networks();
			NetworkSettings[] network_settings = network_array.get_network_data();
			
			String output_text = new String("NETWORK STRUCTURE\n");
			output_text += "Network Type:   " + loaded_network_type + "\n";
			output_text += "Input Type:   " + loaded_input_type + "\n\n\n";
			for (int i = 0; i < network_contents.length; i++)
			{
				output_text += "NETWORK " + i + "\n\n";
				output_text += "Learning Rate: " + network_settings[i].learning_rate + "\n";
				output_text += "Momentum: " + network_settings[i].momentum + "\n";
				output_text += "Input Nodes: " + network_settings[i].num_input_nodes + "\n";
				output_text += "Hidden Nodes: " + network_settings[i].num_hidden_nodes + "\n";
				output_text += "Output Nodes: " + network_settings[i].num_output_nodes + "\n\n";
				output_text += network_contents[i];
				output_text += "\n\n\n";
			}

			text_area_left.setText(output_text);
		}
	}


	class load_input_data_action extends AbstractAction
	// Bring up a JFileChooser dialog box to open a data file containing data to
	// be run through the network. This data may be used for training or may
	// just be data to be processed.
	{
		public void actionPerformed(ActionEvent event)
		{
			int loader_return = data_loader.showOpenDialog(ArrayOfNetworksGUI.this);
			if (loader_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				// Set all dialog boxes to this directory
				network_loader.setCurrentDirectory(data_loader.getCurrentDirectory());
				saver.setCurrentDirectory(data_loader.getCurrentDirectory());
				MIDI_saver.setCurrentDirectory(data_loader.getCurrentDirectory());

				DataIOSystem reader = new DataIOSystem();
				try
				{
					// Read input data into input_data array
					input_data = reader.get_input_data(data_loader.getSelectedFile().getPath(),
					                                   network_array.length,
					                                   network_array.get_network_data(),
					                                   loaded_network_type,
					                                   loaded_input_type,
					                                   loaded_number_bins_partials,
					                                   loaded_number_instruments,
					                                   loaded_number_notes);

					// Get any additional useful data from the input data
					additional_info = reader.get_additional_input_info(data_loader.getSelectedFile().getPath(),
					                                                   network_array.length,
					                                                   network_array.get_network_data());

					// Update what user is permitted to do
					display_input_data_button.setEnabled(true);
					run_samples_button.setEnabled(true);
					load_model_output_button.setEnabled(true);
					display_model_output_button.setEnabled(false);
					save_midi_file_button.setEnabled(false);
					
					// Display success message
					text_area_left.setText("Input data loaded succesfully.\n");
					text_area_centre.setText("");
					text_area_right.setText("");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Inappropriate input data file specified.\n");
				}				
			}
		}
	}


	class display_input_action extends AbstractAction
	// Display the loaded input data in the middle display area
	{
		public void actionPerformed(ActionEvent event)
		{
			text_area_centre.setText("LOADED INPUT DATA");
			for (int i = 0; i < input_data.length; i++)
			{
				text_area_centre.append("\n\n\nNEURAL NETWORK " + i + "\n\n");
				for (int j = 0; j < input_data[i].length; j++)
				{
					text_area_centre.append("Network " + i + "   Input Set " + j + "\n");
					for (int k = 0; k < input_data[i][j].length; k++)
					{
						text_area_centre.append("    Input Node: " + k + "   Value: " +
										input_data[i][j][k] + "\n");
					}
				}
			}
		}
	}


	class load_model_output_action extends AbstractAction
	// Bring up a JFileChooser dialog box to open a data file containing model output
	// data that the networks can be trained on.
	{
		public void actionPerformed(ActionEvent event)
		{
			int loader_return = data_loader.showOpenDialog(ArrayOfNetworksGUI.this);
			if (loader_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				// Set all dialog boxes to this directory
				network_loader.setCurrentDirectory(data_loader.getCurrentDirectory());
				saver.setCurrentDirectory(data_loader.getCurrentDirectory());
				MIDI_saver.setCurrentDirectory(data_loader.getCurrentDirectory());

				DataIOSystem reader = new DataIOSystem();
				try
				{
					// Include number instruments, min pitch and max pitch in additional info
					// Also include rate at which to sample MIDI data
					double sampling_rate = 1000.0 / ((Double) additional_info).doubleValue();
					MIDIConfiguration additional_info = new MIDIConfiguration
					(
						sampling_rate,
						input_data[0].length,
						Integer.parseInt(number_instruments_field.getText()),	// WARNING: will be problems if user changes these 3
						Integer.parseInt(lowest_pitch_field.getText()),
						Integer.parseInt(highest_pitch_field.getText())
					);

					// Read output data into output_data array
					output_data = reader.get_model_output_data(data_loader.getSelectedFile().getPath(),
													           network_array.length,
													           network_array.get_network_data(),
					                                           additional_info,
					                                           loaded_network_type,
					                                           loaded_number_instruments,
					                                           loaded_number_notes);

					/* FOR DEBUGGING PURPOSES
					System.out.println("--- " + sampling_rate);
					for (int i = 0; i < output_data.length; i++)
					{
						System.out.println("NETWORK " + i);
						for (int j = 0; j< output_data[i].length; j++)
						{
							for (int k = 0; k < output_data[i][j].length; k++)
							{
								System.out.print(output_data[i][j][k]);
							}
						}
						System.out.print("\n\n");
					}
					*/

					// Update what user is permitted to do
					train_network_button.setEnabled(true);
					display_model_output_button.setEnabled(true);
					save_midi_file_button.setEnabled(false);

					// Display success message
					text_area_left.setText("Model output data loaded succesfully.\n");
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Inappropriate model output data file specified.\n");
				}				
			}
		}
	}


	class display_model_output_action extends AbstractAction
	// Display the loaded model output data in the middle display area
	{
		public void actionPerformed(ActionEvent event)
		{
			text_area_centre.setText("LOADED MODEL OUTPUT DATA");
			for (int i = 0; i < output_data.length; i++)
			{
				text_area_centre.append("\n\n\nNEURAL NETWORK " + i + "\n\n");
				for (int j = 0; j < output_data[i].length; j++)
				{
					text_area_centre.append("Network " + i + "   Input Set " + j + "\n");
					for (int k = 0; k < output_data[i][j].length; k++)
					{
						text_area_centre.append("    Output Node: " + k + "   Value: " +
										output_data[i][j][k] + "\n");
					}
				}
			}
		}
	}


	class train_network_action extends AbstractAction
	// Trains the networks using the loaded input data and expected output data.
	// Displays the sum-of-squared-errors for every 100th epoch in the right display area.
	{
		public void actionPerformed(ActionEvent event)
		{
			// Randomize training input data if necessary
			double[][][] temp_input_data;
			double[][][] temp_output_data;
			if (randomize_training_checkbox.isSelected())
			{
				temp_input_data = new double[ input_data.length ][ input_data[0].length ][ input_data[0][0].length ];
				temp_output_data = new double[ output_data.length ][ output_data[0].length ][ output_data[0][0].length ];
				boolean[] chosen = new boolean[input_data[0].length];
				for (int i = 0; i < chosen.length; i++)
					chosen[i] = false;
				for (int i = 0; i < input_data[0].length; i++)
				{
					int index = (int) (Math.random() * ((double) input_data[0].length));
					while (chosen[index] == true) // find an index that has not been chosen yet
					{
						index++;
						if (index == chosen.length)
							index = 0;
					}
					chosen[index] = true;
					for (int network = 0; network < input_data.length; network++)
					{
						temp_input_data[network][i] = input_data[network][index];
						temp_output_data[network][i] = output_data[network][index];
					}
				}
			}
			else
			{
				temp_input_data = input_data;
				temp_output_data = output_data;
			}


			// Train network and produce array of the average sum-of-squared-error for
			// each training iteration of each network
			double[][] error_values = network_array.train_networks
			(
				Integer.parseInt(number_iterations_field.getText()),
				temp_input_data,
				temp_output_data
			);

			// Limit what user can do
			save_midi_file_button.setEnabled(false);

			// Print average sum-of-squared error for every 100th epoch of each network
			Double val; // used for rounding for display purposes
			text_area_right.setText("AVERAGE SUM-OF-SQUARED ERRORS\n");
			for (int i = 0; i < error_values.length; i++)
			{
				text_area_right.append("\n\nNETWORK " + i + "\n");
				for (int j = 0; j < error_values[i].length; j++)
				{
					if (j % 100 == 0)
					{
						// Print to display
						val = new Double(error_values[i][j]);
						text_area_right.append("\n   Epoch: " + j + "   Av. Error: " +
						              val.floatValue());
					}
				}
			}
		}
	}


	class run_samples_action extends AbstractAction
	// Displays the results of passing the loaded input data through the network
	// in the central display area. The results that are displayed are in the
	// form usable to the user, not the raw form stored in output_data.
	{
		public void actionPerformed(ActionEvent event)
		{
			// Get unformatted array of outputs
			double[][][] results = network_array.get_outputs(input_data);

			// Ensure that only one voice is playing at a time per instrument if this option is selected
			if (instrument_monophony_checkbox.isSelected())
			{
				if (loaded_network_type.equals("a net/note/inst"))
				{
					for (int dataset = 0; dataset < results[0].length; dataset++)
					{
						for (int instrument = 0; instrument < loaded_number_instruments; instrument++)
						{
							int highest_index = 0;
							double highest_value = 0;

							// Find note for the instrument with highest liklihood
							for (int i = 0; i < loaded_number_notes; i++)
							{
								if (results[ i + (instrument * loaded_number_notes) ][dataset][0] >= highest_value)
								{
									highest_index = i + (instrument * loaded_number_notes);
									highest_value = results[ i + (instrument * loaded_number_notes) ][dataset][0];
								}
							}

							// Set all notes to 0 but the likliest
							for (int i = 0; i < loaded_number_notes; i++)
								if (i + (instrument * loaded_number_notes) != highest_index)
									results[ i + (instrument * loaded_number_notes) ][dataset][0] = 0;
						}
					}
				}
				else if (loaded_network_type.equals("a net/inst"))
				{
					for (int network = 0; network < results.length; network++)
					{
						for (int dataset = 0; dataset < results[0].length; dataset++)
						{
							int highest_index = 0;
							double highest_value = 0;

							// Find note for the instrument with highest liklihood
							for (int output = 0; output < results[0][0].length; output++)
							{
								if (results[network][dataset][output] >= highest_value)
								{
									highest_index = 0;
									highest_value = results[network][dataset][output];
								}
							}
						
							// Set all notes to 0 but the likliest
							for (int output = 0; output < results[0][0].length; output++)
								if (output != highest_index)
									results[network][dataset][output] = 0;
						}
					}
				}
				else if (loaded_network_type.equals("a combined net"))
				{
					for (int dataset = 0; dataset < results[0].length; dataset++)
					{
						for (int instrument = 0; instrument < loaded_number_instruments; instrument++)
						{
							int highest_index = 0;
							double highest_value = 0;

							// Find note for the instrument with highest liklihood
							for (int i = 0; i < loaded_number_notes; i++)
							{
								if (results[0][dataset][ i + (instrument * loaded_number_notes) ] >= highest_value)
								{
									highest_index = i + (instrument * loaded_number_notes);
									highest_value = results[0][dataset][ i + (instrument * loaded_number_notes) ];
								}
							}

							// Set all notes to 0 but the likliest
							for (int i = 0; i < loaded_number_notes; i++)
								if (i + (instrument * loaded_number_notes) != highest_index)
									results[0][dataset][ i + (instrument * loaded_number_notes) ] = 0;
						}
					}
				}				
				else if (loaded_network_type.equals("a limited net"))
				{
					for (int dataset = 0; dataset < results[0].length; dataset++)
					{
						int highest_index = 0;
						double highest_value = 0;

						// Find note for the instrument with highest liklihood
						for (int output = 0; output < results[0][0].length; output++)
						{
							if (results[0][dataset][output] >= highest_value)
							{
								highest_index = output;
								highest_value = results[0][dataset][output];
							}
						}

						// Set all notes to 0 but the likliest
						for (int output = 0; output < results[0][0].length; output++)
							if (output != highest_index)
								results[0][dataset][output] = 0;
					}
				}
			}

			// Translate results
			DataIOSystem interpreter = new DataIOSystem();
			double[][][] interpreted_results = interpreter.interpret_results(results);

			// Clean results if option is selected
			// Turn all 101 to 111 and 010 to 000
			if (clean_output_checkbox.isSelected())
			{
				for (int network = 0; network < interpreted_results.length; network++)
				{
					for (int output = 0; output < interpreted_results[network][0].length; output++)
					{
						for (int dataset = 1; dataset < interpreted_results[network].length - 1; dataset++)
						{
							if (interpreted_results[network][dataset-1][output] > 0.9 && interpreted_results[network][dataset+1][output] > 0.9)
								interpreted_results[network][dataset][output] = 1;
							else if (interpreted_results[network][dataset-1][output] < 0.9 && interpreted_results[network][dataset+1][output] < 0.9)
								interpreted_results[network][dataset][output] = 0;
						}
					}
				}
			}

			// Store the results externally
			stored_results = interpreted_results;

			// Print translated results in right display area
			text_area_right.setText("OUTPUT OF NETWORKS\n");
			Double val; // used for rounding for display purposes
			for (int i = 0; i < interpreted_results.length; i++)
			{
				text_area_right.append("\n\nNEURAL NETWORK " + i + "\n\n");
				for (int j = 0; j < interpreted_results[i].length; j++)
				{
					text_area_right.append("Network " + i + "   Input Set " + j + "\n");
					for (int k = 0; k < interpreted_results[i][j].length; k++)
					{
						val = new Double(interpreted_results[i][j][k]);
						text_area_right.append("    Output Node: " + k + "   Value: " +
										 val.floatValue() + "\n");
					}
				}
			}

			// Print average error per network
			text_area_right.append("\n\nERROR RATE PER NETWORK\n");
			for (int network = 0; network < output_data.length; network++)
			{
				int correct = 0;
				int total = 0;
				text_area_right.append("\n   Network " + network + ": ");
				for (int dataset = 0; dataset < output_data[network].length; dataset++)
				{
					for (int output = 0; output < output_data[network][dataset].length; output++)
					{
						if (Math.abs(output_data[network][dataset][output] - interpreted_results[network][dataset][output]) < 0.01)
							correct++;
						total++;
					}
				}
				int error_rate = (int) (100 * ((float) correct) / ((float) total));
				text_area_right.append(error_rate + "% (" + correct + " / " + total + ")");
			}

			// Print average error per output unit
			int total_correct = 0;
			int complete_total = 0;
			text_area_right.append("\n\nERROR RATE PER OUTPUT UNIT\n");
			int total = output_data[0].length;
			for (int network = 0; network < output_data.length; network++)
			{
				text_area_right.append("\n   Network " + network + ": ");
				for (int output = 0; output < output_data[network][0].length; output++)
				{
					int correct = 0;
					text_area_right.append("\n      Output " + output + ": ");
					for (int dataset = 0; dataset < output_data[network].length; dataset++)
					{
						if (Math.abs(output_data[network][dataset][output] - interpreted_results[network][dataset][output]) < 0.01)
						{
							correct++;
							total_correct++;
						}
						complete_total++;
					}
					int error_rate = ((int) (100.0 * ((float) correct) / ((float) total)));
					text_area_right.append(error_rate + "% (" + correct + " / " + total + ")");
				}
			}

			// Print overall average
			text_area_right.append("\n\nOVERALL AVERAGE ERROR RATET\n");
			float average_error = (float) (100.0 * ((double) total_correct) / ((double) complete_total));
			text_area_right.append("   " + average_error + "% (" + total_correct + " / " + complete_total + ")\n");

			// Allow user to save results
			save_midi_file_button.setEnabled(true);
		}
	}


	class save_midi_action extends AbstractAction
	// Save the transcription results in a MIDI file
	{
		public void actionPerformed(ActionEvent event)
		{
			int saver_return = MIDI_saver.showSaveDialog(ArrayOfNetworksGUI.this);
			if (saver_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				// Set all dialog boxes to this directory
				network_loader.setCurrentDirectory(MIDI_saver.getCurrentDirectory());
				data_loader.setCurrentDirectory(MIDI_saver.getCurrentDirectory());
				saver.setCurrentDirectory(MIDI_saver.getCurrentDirectory());

				MIDISequenceWriter midi_writer = new MIDISequenceWriter();
				File selection = MIDI_saver.getSelectedFile();
				try
				{
					String MIDI_filename = selection.getPath();
					if (loaded_network_type.equals("instrument ID"))
						text_area_left.setText("ERROR:\n" + "Cannot save instrument ID as MIDI.\n");
					else
					{
						midi_writer.save_as_MIDI(loaded_number_instruments, ((Double) additional_info).doubleValue(), loaded_min_note,
												loaded_max_note, stored_results, loaded_network_type, MIDI_filename);
						text_area_left.setText("MIDI file saved succesfully.\n");
					}
				}
				catch (Exception e)
				{
					// e.printStackTrace();
					text_area_left.setText("ERROR:\n" + "Unable to save MIDI file.\n");
				}
			}
		}
	}
}