/*
DataIOSystem
by Cory McKay

CLASS NAME:	NetworksFileReader
CLASS TYPE:	File reader and processor
DESCRIPTION:	Reads in input files and expected output files that can be processed through
			neural networks and can be used to train the networks. Although the other classes
			of this system are designed to be multipurpose, the following methods of this
			class may need to be modified to meet the format of input and output files
			that will be used for particular applicatoins:
				get_additional_input_info
				interpret_results
				parse_input_data
				parse_model_output_data
				scale_input_data
				scale_model_output_data
			Objects of this class also translate raw network output data into a meaningufl
			form (using the interpret_results method).
			The get_input_data method can be called to a load a file containing input data
			for a network. This data can be used to train networks or can be just processed
			by the network.
			The get_additional_input_info method can be called after the get_input_data to get
			any additional useful information after the input data has been called. This will
			not always be necessary, in which case this method should be set to return null.
			The get_model_output_data method can be called to load a file containg model output
			that corresponds to the loaded input data. This is used for training the networks.
*/

import java.io.*;
import java.util.StringTokenizer;

class DataIOSystem
{
	/* FIELDS ******************************************************************/


	private int next_index;	// the current location in the chars array
	private char[] chars;	// an array that holds all of the characters that are read in through
							//     the file


	/* PUBLIC METHODS **********************************************************/


	public double[][][] get_input_data(String filename, int number_networks, NetworkSettings[] network_data,
                                       String loaded_network_type, String loaded_input_type, int number_bins_partials,
	                                   int number_instruments, int number_notes)
		throws IOException
	// Parses the given file to correspond with the given network configurations.
	// Scales the results so that all inputs fall between 0.25 and 0.75.
	// The first index of the specified array specifies the network index, the
	// second specifies data set number and the third specifies input node.
	{
		// Initialize fields
		chars = null;
		next_index = 0;

		// Read input data file into chars
		access_file(filename);

		// Parse the contents of the file
		double[][][] input_data = parse_input_data(number_networks, network_data,
		                                           loaded_network_type, loaded_input_type, 
		                                           number_bins_partials, number_instruments,
		                                           number_notes);

		/* Audio analysis system does this now
		// Scale input data to 0.25 to 0.75 range (sigmoidal activation function)
		scale_input_data(input_data, number_networks, network_data);
		*/

		// Reset fields
		chars = null;
		next_index = 0;

		// Return scaled input data
		return input_data;
	}


	public Object get_additional_input_info(String filename, int number_networks,
	                                        NetworkSettings[] network_data)
		throws IOException
	// Return any additional information from the input data file that might be useful
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		// Initialize fields
		chars = null;
		next_index = 0;

		// Read input data file into chars
		access_file(filename);

		// Parse the contents of the file
		String duration_line = find_line_starting_with("FrameTime(ms)");
		StringTokenizer	tokens = new StringTokenizer(duration_line, " ");
		String temp = tokens.nextToken();
		Object duration = new Double(tokens.nextToken());

		// Reset fields
		chars = null;
		next_index = 0;

		// Return duration of one frame in milliseconds
		return duration;
	}


	public double[][][] get_model_output_data(String filename, int number_networks,
	                                          NetworkSettings[] network_data, MIDIConfiguration additional_info,
	                                          String network_type, int number_instruments, int number_notes)
		throws Exception
	// Parses the given file to correspond with the given network configurations.
	// Extracts model output data from the file that can be used to train an array of networks.
	// The first index of the retruned array specifies the network index, the
	// second specifies data set number and the third specifies output node.
	{
		int min_pitch = additional_info.min_pitch;
		int number_datasets = additional_info.total_samples;
		
		// Extract score from MIDI file. The 1st indice of the score corresponds to channel, the 2nd tosample
		// number and the 3rd corresponds to the MIDI pitch. Entries are marked true when a given pitch
		// is playing at a given sample.
		MIDIFileReader reader = new MIDIFileReader(filename, additional_info.sampling_rate, number_instruments);
		boolean[][][] score = reader.getScore();

		// Prepare array to be returned. 1 denotes that a pitch is playing, 0 denotes that
		// it is not.
		double[][][] return_data = new double[number_networks][number_datasets][];
		if (network_type.equals("a net/note/inst"))
		{
			for (int instrument = 0; instrument < number_instruments; instrument++)
			{
				for (int note = 0; note < number_notes; note++)
				{
					for (int dataset = 0; dataset < number_datasets; dataset++)
					{
						int net_index = instrument * number_notes + note;
						return_data[net_index][dataset] = new double[1];
						if (score[instrument][dataset][min_pitch + note] == true)
							return_data[net_index][dataset][0] = 1;
						else
							return_data[net_index][dataset][0] = 0;
					}
				}
			}
		}
		else if (network_type.equals("a net/inst"))
		{
			for (int instrument = 0; instrument < number_instruments; instrument++)
			{
				for (int dataset = 0; dataset < number_datasets; dataset++)
				{
					return_data[instrument][dataset] = new double[number_notes];
					for (int note = 0; note < number_notes; note++)
					{
						if (score[instrument][dataset][min_pitch + note] == true)
							return_data[instrument][dataset][note] = 1;
						else
							return_data[instrument][dataset][note] = 0;
					}
				}
			}
		}
		else if (network_type.equals("a combined net"))
		{
			for (int dataset = 0; dataset < number_datasets; dataset++)
			{
				return_data[0][dataset] = new double[number_instruments * number_notes];
				for (int instrument = 0; instrument < number_instruments; instrument ++)
				{
					for (int note = 0; note < number_notes; note++)
					{
						int output_index = instrument * number_notes + note;
						if (score[instrument][dataset][min_pitch + note] == true)
							return_data[0][dataset][output_index] = 1;
						else
							return_data[0][dataset][output_index] = 0;
					}
				}
			}
		}
		else if (network_type.equals("a limited net"))
		{
			for (int dataset = 0; dataset < number_datasets; dataset++)
			{
				return_data[0][dataset] = new double[number_notes];
				for (int note = 0; note < number_notes; note++)
				{
					for (int instrument = 0; instrument < number_instruments; instrument++)
					{
						if (score[instrument][dataset][min_pitch + note] == true)
						{
							return_data[0][dataset][note] = 1;
							instrument = number_instruments; // exit the loop
						}
						else
							return_data[0][dataset][note] = 0;
					}
				}
			}
		}
		else if (network_type.equals("instrument ID"))
		{
			for (int dataset = 0; dataset < number_datasets; dataset++)
			{
				return_data[0][dataset] = new double[number_instruments];
				for (int instrument = 0; instrument < number_instruments; instrument++)
				{
					for (int note = 0; note < 128; note++)
					{
						if (score[instrument][dataset][note] == true)
						{
							return_data[0][dataset][instrument] = 1;
							note = 128; // exit the loop
						}
						else
							return_data[0][dataset][instrument] = 0;
					}
				}
			}
		}
		
		return return_data;
	}


	public double[][][] interpret_results(double[][][] raw_results)
	// Returns an array holding a translation of the raw output from network into a form
	// useful for the application at hand.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		double[][][] translation = new double[raw_results.length][][]; // translated results
		
		// In this case, set every result below 0.5 to 0 and and every result above it to 1
		for (int i = 0; i < raw_results.length; i++)
		{
			translation[i] = new double[raw_results[i].length][];
			for (int j = 0; j < raw_results[i].length; j++)
			{
				translation[i][j] = new double[raw_results[i][j].length];
				for (int k = 0; k < raw_results[i][j].length; k++)
				{
					if (raw_results[i][j][k] < 0.5) translation[i][j][k] = 0;
					else translation[i][j][k] = 1;
				}
			}
		}

		return translation;
	}


	/* PRIVATE METHODS *********************************************************/

	
	private void access_file(String filename) 
		throws IOException
	// Returns the contents of the specified file in the chars array
	{
		File src = new File(filename);
		chars = new char[(int) src.length()];
		src = null;
		
		InputStream in;
		in = new FileInputStream(filename);
		int ch;
		for (int i = 0; (ch = in.read()) != -1; i++)
			chars[i] = (char)ch;
		in.close();
	}


	private String get_next_string()
	// Reads in one line from chars and returns it as a string. Also moves next_index to
	// the next_location in the file.
	{
		int i = next_index;
		int done = -1;

		while (done != 1)
		{
			if (i < chars.length && done != 1)
			{
				if ( chars[i] != '\n' )
					i++;
				else done = 1;
			}
			else done = 1;
		}
		String temp = new String(chars, next_index, i-next_index);
		next_index = i + 1;
		return temp;
	}


	private String find_line_starting_with(String search_string)
	// Return the first line from the current location in chars that starts with search_string
	{
		String current_line = null; // the current line of information being read
		StringTokenizer tokens; // used to parse each line of information
		boolean done = false; // keeps track of when loop is finished
		while (done == false)
		{
			current_line = get_next_string();		
			tokens = new StringTokenizer(current_line, " ");
			if (tokens.nextToken().equals(search_string))
				done = true;
		}
		return current_line;
	}


	private double[][][] parse_input_data(int number_networks, NetworkSettings[] network_data,
	                                      String network_type, String input_type,
	                                      int number_bins_partials, int number_instruments,
	                                      int number_notes)
	// Parses the data that has been read into chars and that corresponds to the given
	// network. Returns it in as an array whose first index specifies the network index,
	// second specifies data set number and the third specifies input node.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		// Set up variables
		double[][][] data; // data to be returned
		int number_data_sets = 1; // the total number of data sets in the data
		int lines_per_data_set = 1; // the number of lines per data set
		double[][] freq_data = new double[1][1]; // used to temporarily store the frequency data being read
		double[][] ampl_data = new double[1][1]; // used to temporarily store the amplitude data being read
		String current_line = null; // the current line of information being read
		StringTokenizer tokens; // used to parse each line of information

		// Find the number of data sets in the input data
		current_line = find_line_starting_with("NumberOfFrames");
		current_line = current_line.substring(0, (current_line.length() - 1));
		tokens = new StringTokenizer(current_line, " ");
		current_line = tokens.nextToken();
		current_line = tokens.nextToken();
		number_data_sets = Integer.parseInt(current_line);

		// Find the number of lines per data set
		current_line = get_next_string();		
		current_line = current_line.substring(0, (current_line.length() - 1));
		tokens = new StringTokenizer(current_line, " ");
		current_line = tokens.nextToken();
		lines_per_data_set = Integer.parseInt(tokens.nextToken());

		// Parse file data into freq_data and possible ampl_data
		freq_data = new double[number_data_sets][number_bins_partials];
		ampl_data = new double[number_data_sets][number_bins_partials];
		for (int set = 0; set < number_data_sets; set++)
		{
			// Read in irrelevant lines
			current_line = get_next_string();
			current_line = get_next_string();

			// Go through one data set
			for (int i = 0; i < lines_per_data_set; i++)
			{
				current_line = get_next_string();
				current_line = current_line.substring(0, (current_line.length() - 1));
				tokens = new StringTokenizer(current_line, " ");
				if (i < number_bins_partials)
				{
					freq_data[set][i] = Double.parseDouble(tokens.nextToken());
					if (input_type.equals("f and A"))
						ampl_data[set][i] = Double.parseDouble(tokens.nextToken());
				}
			}
		}

		// Prepare an array to hold the data to be returned
		data = new double[number_networks][number_data_sets][network_data[0].num_input_nodes]; // assumes all networks have same # input nodes

		// Copy data from freq_data and possibly ampl_data into data
		if (network_type.equals("a net/note/inst"))
		{
			for (int instrument = 0; instrument < number_instruments; instrument++)
			{
				for (int note = 0; note < number_notes; note++)
				{
					for (int dataset = 0; dataset < number_data_sets; dataset++)
					{
						for (int input = 0; input < number_bins_partials; input++)
						{
							int net_index = instrument * number_notes + note;
							if (input_type.equals("f and A"))
							{
								data[net_index][dataset][input * 2] = freq_data[dataset][input];
								data[net_index][dataset][(input * 2) + 1] = ampl_data[dataset][input];
							}
							else
								data[net_index][dataset][input] = freq_data[dataset][input];
						}
					}
				}
			}
		}
		else if (network_type.equals("a net/inst"))
		{
			for (int instrument = 0; instrument < number_instruments; instrument++)
			{
				for (int dataset = 0; dataset < number_data_sets; dataset++)
				{
					for (int input = 0; input < number_bins_partials; input++)
					{
						if (input_type.equals("f and A"))
						{
							data[instrument][dataset][input * 2] = freq_data[dataset][input];
							data[instrument][dataset][(input * 2) + 1] = ampl_data[dataset][input];
						}
						else
							data[instrument][dataset][input] = freq_data[dataset][input];
					}
				}
			}
		}
		else if (network_type.equals("a combined net") || network_type.equals("a limited net") || network_type.equals("instrument ID"))
		{
			for (int dataset = 0; dataset < number_data_sets; dataset++)
			{
				for (int input = 0; input < number_bins_partials; input++)
				{
					if (input_type.equals("f and A"))
					{
						data[0][dataset][input * 2] = freq_data[dataset][input];
						data[0][dataset][(input * 2) + 1] = ampl_data[dataset][input];
					}
					else
						data[0][dataset][input] = freq_data[dataset][input];
				}
			}
		}

		// Return the input data
		return data;
	}


	private void scale_input_data(double[][][] data, int number_networks,
	                              NetworkSettings[] network_data)
	// Scale the input data so that it is of a convenient size to serve as input to
	// the networks.
	// THIS METHOD WILL NEED TO BE MODIFIED TO FIT THE NEEDS OF ANY GIVEN PROJECT
	{
		for (int i = 0; i < data.length; i++)
		{
			for (int j = 0; j < data[i].length; j++)
			{
				for (int k = 0; k < data[i][j].length; k++)
					data[i][j][k] = (data[i][j][k] / 12000) + 0.25;
			}
		}
	}
}