/*
ExceptionSettingsDialog
by Cory McKay

CLASS NAME:		ExceptionSettingsDialog
CLASS TYPE:		GUI - dialog box
DESCRIPTION:	This dialog box is used to get the basic settings for a new neural network.
				The user is able to set the number of input, hidden and output nodes, the
				learning rate, the momentum, the minimum initial weight and the maximum
				initial weight for the network.
				The was_updated method can be used to see if the OK button was pressed (as
				opposed to the Cancel button or the close box). The get_settings method
				can be called to get a NetworkSetings object with the selections in this
				dialog box.
*/


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


class ExceptionSettingsDialog
	extends JDialog
{
	/* FIELDS ******************************************************************/


	// Fields holding network configuration data
	public JTextField num_inputs_field;
	public JTextField num_hiddens_field;
	public JTextField num_outputs_field;
	public JTextField learning_rate_field;
	public JTextField momentum_factor_field;
	public JTextField min_init_weight_field;
	public JTextField max_init_weight_field;

	// Set to true after the OK button is pressed, is false otherwise
	private boolean update; 


	/* CONSTRUCTOR *************************************************************/


	public ExceptionSettingsDialog(JDialog owner, int network_index)
	{
		// Give the dialog box its owner, its title and make it modal
		super(owner, "Settings for Network " + network_index, true);

		// Initialize the dialog box as not having been updated (to distinguish between ok and cancel)
		update = false;

		// Prepare panels to display dialog box contents
		Container contentPane = getContentPane();
		JPanel outer_panel = new JPanel();
		outer_panel.setLayout(new BorderLayout());
		JPanel text_panel = new JPanel();
		text_panel.setLayout(new GridLayout(7,2));
		JPanel button_panel = new JPanel();

		// Configure an OK button to close the dialog box
		// Sets update to true so that the calling class can check to see if OK was pressed
		JButton ok = new JButton("OK");
		ok.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				update = true;
				setVisible(false);
			}
		});

		// Configure a cancel button to close the dialog
		JButton cancel = new JButton("Cancel");
		cancel.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				setVisible(false);
			}
		});

		// Initialize text fields
		num_inputs_field = new JTextField("3", 3);
		num_hiddens_field = new JTextField("4", 3);
		num_outputs_field = new JTextField("1", 3);
		learning_rate_field = new JTextField("0.1", 4);
		momentum_factor_field = new JTextField("0.1", 4);
		min_init_weight_field = new JTextField("0.3", 4);
		max_init_weight_field = new JTextField("0.7", 4);

		// Prevent user from editing the number of output nodes directly
		num_outputs_field.setEditable(false);

		// Set up text_panel
		text_panel.add(new JLabel("Inputs Nodes:"));
		text_panel.add(num_inputs_field);
		text_panel.add(new JLabel("Hidden Nodes:"));
		text_panel.add(num_hiddens_field);
		text_panel.add(new JLabel("Outputs Noes:"));
		text_panel.add(num_outputs_field);
		text_panel.add(new JLabel("Learning Rate:"));
		text_panel.add(learning_rate_field);
		text_panel.add(new JLabel("Momentum:"));
		text_panel.add(momentum_factor_field);
		text_panel.add(new JLabel("Min Init Weight:"));
		text_panel.add(min_init_weight_field);
		text_panel.add(new JLabel("Max Init Weight:"));
		text_panel.add(max_init_weight_field);

		// Set up button_panel
		button_panel.add(ok);
		button_panel.add(cancel);

		// Display dialog box contents
		outer_panel.add(text_panel, BorderLayout.CENTER);
		outer_panel.add(button_panel, BorderLayout.SOUTH);
		contentPane.add(outer_panel);
		setSize(220, 220);
	}


	/* PUBLIC METHODS **********************************************************/


	public boolean was_updated()
	// Returns true if the OK button was pressed the last time the dialog box was visible.
	// Returns false otherwise. Resets the update field to false.
	{
		boolean temp = update;
		update = false;
		return temp;
	}


	/***************************************************************************/


	public NetworkSettings get_settings()
	// Returns entered settings
	{
		return new NetworkSettings
		(
			Double.parseDouble(learning_rate_field.getText()),
			Double.parseDouble(momentum_factor_field.getText()),
			Integer.parseInt(num_inputs_field.getText()),
			Integer.parseInt(num_hiddens_field.getText()),
			Integer.parseInt(num_outputs_field.getText()),
			Double.parseDouble(min_init_weight_field.getText()),
			Double.parseDouble(max_init_weight_field.getText())
		);
	}
}