/*
MIDIFileReader
by Cory McKay

Class is constructed given the path of a MIDI file and sampling rate that will be used to
quantize the data. A 2-D boolean array called score is constructed with the first index
reflecting sample number and the second reflecting MIDI pitch. An entry is marked true
if the given pitch is currently playing at the current sample. All MIDI messages except
for note ons and note offs are ignored.

The getScore method can be called to receive the score array. The printScore method
can be called to print (to standard out) a table representing the score (this is used
only for debugging purposes). The public fields can be accessed in order to receive timing
information.

WARNING: tempo change MIDI data in the MIDI file is ignored (not significant for purposes
         of this class, since measuring time based on MIDI ticks, not beats)
WARNING: MIDI pitch bends are ignored
*/


import  java.io.File;
import  java.io.IOException;
import  javax.sound.midi.MidiSystem;
import  javax.sound.midi.InvalidMidiDataException;
import  javax.sound.midi.Sequence;
import  javax.sound.midi.Track;
import  javax.sound.midi.MidiEvent;
import  javax.sound.midi.MidiMessage;
import  javax.sound.midi.ShortMessage;


public class MIDIFileReader
{
	/* FIELDS ******************************************************************/

	
	private Track[] tracks;				// the tracks in a MIDI file
	private boolean[][][] score;		// the score indicating when pitches are playing, indexed by channel
										//	(1st indice), sample number (2nd indice) and MIDI pitch (3rd indice)
	public double sampling_rate;		// sampling rate, in Hz, that sets the temporal resolution of the score
	public double seconds_per_sample;	// number of seconds corresponding to one sample
	public double ticks_per_sec;		// number of MIDI ticks corresponding to 1 second of score time
	public double ticks_per_sample;		// number of MIDI ticks corresponding to one score sample
	public int score_length;			// length of score in samples


	/* CONSTRUCTOR *************************************************************/


	public MIDIFileReader(String midi_filename, double samplingrate, int number_instruments)
		throws Exception
	// Takes in the path of a MIDI file and tha sampling rate, in Hz, that will be used to quantize it
	// temporally. Fills the score array with true at every sample that a note is playing. Also fills
	// in the public fields holding temporal information.
	{
		// Attempt to load the file specified by midi_filename 
		File midi_file = new File(midi_filename);
		Sequence sequence = null;
		try {sequence = MidiSystem.getSequence(midi_file);}
		catch (InvalidMidiDataException e) {/* e.printStackTrace(); */}
		catch (IOException e) {/* e.printStackTrace(); */}
		if (sequence == null) throw new Exception("ERROR: Invalid file specified.");

		// Load the MIDI file into tracks and 
		tracks = sequence.getTracks();

		// Calculate timing information
		sampling_rate = samplingrate;
		seconds_per_sample = 1.0 / sampling_rate;
		ticks_per_sec = ((double) sequence.getTickLength()) / ((double) sequence.getMicrosecondLength() / 1000000.0);
		ticks_per_sample = ticks_per_sec / sampling_rate;
		if ( ((((double) sequence.getTickLength()) / ticks_per_sample) + 1.0) > (double) Integer.MAX_VALUE)
			throw new Exception("ERROR: MIDI file is too big. Please use smaller file or decrease sampling rate.");
		score_length = (int) ((((double) sequence.getTickLength()) / ticks_per_sample) + 1.0);

		// Initialize the score (only first column)
		score = new boolean[number_instruments][score_length][128];
		for (int i = 0; i < score[0][0].length; i++) // initialize first sample column
			score[0][0][i] = false;

		// Fill the score
		fillScore();
	}


	/* PUBLIC METHODS **********************************************************/


	public boolean[][][] getScore()
	// Returns the filled in score. The first indice corresponds to channel number, the second to sample number (corresponding to the
	// given sampling rate) and the third corresponds to the MIDI pitch. Entries are marked true
	// when a given pitch is playing at a given sample.
	{
		return score;
	}

	
	public void printScore()
	// Prints a table where each row corresponds to a pitch (starting at 0) and each column corresponds
	// to a sample (starting at 0). A | on the table means that the pitch is playing at that sample
	// and a _ means it is not.
	// USED FOR DEBUGGING PURPOSES ONLY
	{
		char tf; // holds t or f
		for (int k = 0; k < score.length; k++)
		{
			for (int i = 0; i < 128; i++)
			{
				for (int j = 0; j < score[0].length; j++)
				{
					if (score[k][j][i] == true) tf = '|';
					else tf = '_';
					System.out.print(tf);
				}
				System.out.print("\n");
			}
			System.out.print("\n");
			System.out.print("\n");
		}
	}


	/* PRIVATE METHODS *********************************************************/

	
	private void fillScore()
	// Fill score so that every entry marked true indicates that a pitch is playing at that
	// sample index and channel.
	{
		int previous_sample_index = 0; // the sample index corresponding to the last MIDI event
		int current_sample_index = 0; // the sample index corresponding to the current MIDI event
		for (int nTrack = 0; nTrack < tracks.length; nTrack++)
		{
			Track track = tracks[nTrack];
			for (int nEvent = 0; nEvent < track.size(); nEvent++)
			{
				// Get the MIDI message corresponding to the next MIDI event and find the MIDI tick
				// that it corresponds to
				MidiEvent event = track.get(nEvent);
				MidiMessage message = event.getMessage();
				long current_tick = event.getTick();

				// Cause score to reflect new note ons or note offs and cause note ons left over from
				// last MIDI event to perpetuate up until newest MIDI event
				previous_sample_index = current_sample_index;
				current_sample_index = (int) (((double) current_tick) / ((double) ticks_per_sample));
				if (message instanceof ShortMessage)
				{
					ShortMessage short_message = (ShortMessage) message;
					int channel = short_message.getChannel();
					if (short_message.getCommand() == 0x80) // note off
					{
						updateScore(score[channel], previous_sample_index, current_sample_index);
						score[channel][current_sample_index][short_message.getData1()] = false;
					}
					else if (short_message.getCommand() == 0x90) // note on
					{
						updateScore(score[channel], previous_sample_index, current_sample_index);
						if (short_message.getData2() == 0)
							score[channel][current_sample_index][short_message.getData1()] = false;
						else
							score[channel][current_sample_index][short_message.getData1()] = true;
					}
				}
			}
		}
	}


	private void updateScore(boolean[][] score, int previous_sample_index, int current_sample_index)
	// Fills in the score between the previous_sample_index and current_sample_index so that all pitches
	// that have had note ons but have not had note offs yet are marked as playing on every sample, and
	// all other pitches are marked as not playing.
	{
		// Pitches that were being played at the time of the last MIDI event
		boolean[] previously_playing = new boolean[128];
		for (int i = 0; i < 128; i++)
			previously_playing[i] = score[previous_sample_index][i];

		// Update score
		for (int i = previous_sample_index + 1; i <= current_sample_index; i++)
		{
			for (int j = 0; j < 128; j++)
				score[i][j] = previously_playing[j];
		}
	}
}