/*
MIDISequenceWriter
by Cory McKay

CLASS NAME:		MIDISequenceWriter
CLASS TYPE:		File saver
DESCRIPTION:	Saves transcriptions to MIDI files.
LIMITATION:		Limited by resolution of given frame time
*/


import javax.sound.midi.Sequence;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;
import javax.sound.midi.MidiSystem;
import java.io.File;

class MIDISequenceWriter
{
	/* PUBLIC METHODS **********************************************************/


	public void save_as_MIDI(int number_instruments, double frame_time, int min_pitch, int max_pitch,
	                         double[][][] score_data, String network_type, String file_name)
		throws Exception
	// Save the score_data (the result of processing data through the networks) in a MIDI file
	{
		// Configure the score (first index is instrument, second is dataset)
		// Contents are note on (+ve pitch), (-ve pitch), or neither (0).
		int[][] score = new int[number_instruments][score_data[0].length];
		for (int i = 0; i < score.length; i++)
			for (int j = 0; j < score[i].length; j++)
				score[i][j] = 0;
		int number_notes = max_pitch - min_pitch + 1;
		boolean[] playing = new boolean[number_notes]; // if a current pitch is currently playing
		for (int i = 0; i < playing.length; i++)
			playing[i] = false;
		if (network_type.equals("a net/note/inst"))
		{
			for (int network = 0; network < score_data.length; network++)
			{
				int instrument = network / number_notes;
				int pitch = min_pitch + (network % number_notes);
				for (int dataset = 0; dataset < score_data[0].length; dataset++)
				{
					if (score_data[network][dataset][0] > 0.9 && playing[pitch - min_pitch] == false)
					{
						playing[pitch - min_pitch] = true;
						if (score[instrument][dataset] != 0 && (dataset != score_data[0].length - 1)) // if note off and note on occur simultaneously
							score[instrument][dataset + 1] = pitch; // note on
						else
							score[instrument][dataset] = pitch; // note on
					}
					if (score_data[network][dataset][0] < 0.1 && playing[pitch - min_pitch] == true)
					{
						playing[pitch - min_pitch] = false;
						if (score[instrument][dataset] != 0 && (dataset != 0)) // if note off and note on occur simultaneously
							score[instrument][dataset - 1] = -1 * pitch; // note off
						else
							score[instrument][dataset] = -1 * pitch; // note off
					}
				}
				score[instrument][score_data[0].length - 1] = -1 * pitch; // final note off
			}
		}
		else if (network_type.equals("a net/inst"))
		{
			for (int instrument = 0; instrument < number_instruments; instrument++)
			{
				for (int output = 0; output < number_notes; output++)
				{
					int pitch = min_pitch + output;
					for (int dataset = 0; dataset < score_data[0].length; dataset++)
					{
						if (score_data[instrument][dataset][output] > 0.9 && playing[pitch - min_pitch] == false)
						{
							playing[pitch - min_pitch] = true;
							if (score[instrument][dataset] != 0) // if note off and note on occur simultaneously
								score[instrument][dataset + 1] = pitch; // note on
							else
								score[instrument][dataset] = pitch; // note on
						}
						if (score_data[instrument][dataset][output] < 0.1 && playing[pitch - min_pitch] == true)
						{
							playing[pitch - min_pitch] = false;
							if (score[instrument][dataset] != 0) // if note off and note on occur simultaneously
								score[instrument][dataset - 1] = -1 * pitch; // note off
							else
								score[instrument][dataset] = -1 * pitch; // note off
						}
					}
					score[instrument][score_data[0].length - 1] = -1 * pitch; // final note off
				}
			}
		}
		else if (network_type.equals("a combined net"))
		{
			for (int instrument = 0; instrument < number_instruments; instrument++)
			{
				for (int note = 0; note < number_notes; note++)
				{
					int output = note + (instrument * number_notes);
					int pitch = min_pitch + note;
					for (int dataset = 0; dataset < score_data[0].length; dataset++)
					{
						if (score_data[0][dataset][output] > 0.9 && playing[pitch - min_pitch] == false)
						{
							playing[pitch - min_pitch] = true;
							if (score[instrument][dataset] != 0) // if note off and note on occur simultaneously
								score[instrument][dataset + 1] = pitch; // note on
							else
								score[instrument][dataset] = pitch; // note on
						}
						if (score_data[0][dataset][output] < 0.1 && playing[pitch - min_pitch] == true)
						{
							playing[pitch - min_pitch] = false;
							if (score[instrument][dataset] != 0) // if note off and note on occur simultaneously
								score[instrument][dataset - 1] = -1 * pitch; // note off
							else
								score[instrument][dataset] = -1 * pitch; // note off
						}
					}
					score[instrument][score_data[0].length - 1] = -1 * pitch; // final note off
				}
			}
		}
		else if (network_type.equals("a limited net"))
		{
			int instrument = 0;
			for (int note = 0; note < number_notes; note++)
			{
				int pitch = min_pitch + note;
				for (int dataset = 0; dataset < score_data[0].length; dataset++)
				{
					if (score_data[0][dataset][note] > 0.9 && playing[pitch - min_pitch] == false)
					{
						playing[pitch - min_pitch] = true;
						if (score[instrument][dataset] != 0) // if note off and note on occur simultaneously
							score[instrument][dataset + 1] = pitch; // note on
						else
							score[instrument][dataset] = pitch; // note on
					}
					if (score_data[0][dataset][note] < 0.1 && playing[pitch - min_pitch] == true)
					{
						playing[pitch - min_pitch] = false;
						if (score[instrument][dataset] != 0) // if note off and note on occur simultaneously
							score[instrument][dataset - 1] = -1 * pitch; // note off
						else
							score[instrument][dataset] = -1 * pitch; // note off
					}
				}
				score[instrument][score_data[0].length - 1] = -1 * pitch; // final note off
			}
		}

		// Create the sequencer
		int resolution = (int) (40.0 / frame_time); // ticks per frame (40 because 25 frames per sec)
		if (resolution == 0) resolution = 1;
		Sequence seq = new Sequence(Sequence.SMPTE_25, resolution, 2);

		// Get the tracks
		Track[] tracks = new Track[number_instruments];
		for (int i = 0; i < number_instruments; i++)
			tracks[i] = seq.createTrack();

		// Add MIDI events to the tracks
		for (int instrument = 0; instrument < number_instruments; instrument++)
		{
			// Select the instrument
			ShortMessage message = new ShortMessage();
			if (instrument == 0) // Clarinet
				message.setMessage(ShortMessage.PROGRAM_CHANGE, instrument, 71, 0); // clarinet
			else // Bassoon
				message.setMessage(ShortMessage.PROGRAM_CHANGE, instrument, 70, 0); // bassoon
			MidiEvent event = new MidiEvent(message, 0);
			tracks[instrument].add(event);
			
			// Add note ons and note offs
			for (int dataset = 0; dataset < score[instrument].length; dataset++)
			{
				if (score[instrument][dataset] > 0)
				{
					add_MIDI_note(dataset, frame_time, tracks, instrument, score[instrument][dataset], 100); // note on
				}
				else if (score[instrument][dataset] < 0)
				{
					add_MIDI_note(dataset, frame_time, tracks, instrument, -1 * score[instrument][dataset], 0); // note off
				}
			}
		}

		// Save the file
		MidiSystem.write(seq, 1, new File(file_name));
	}


	/* PRIVATE METHODS *********************************************************/


	private static void add_MIDI_note(int dataset, double frame_time, Track[] tracks, int instrument, int pitch, int velocity)
		throws Exception
	// Add a note on to the appropriate place in tracks
	{
		ShortMessage message = new ShortMessage();
		message.setMessage(ShortMessage.NOTE_ON, instrument, pitch, velocity) ;
		long tick = (long) (((double) dataset) * frame_time / 40.0);
		MidiEvent event = new MidiEvent(message, tick);
		tracks[instrument].add(event);
	}
}