/*
NNController
by Cory McKay

CLASS NAME:		NNController
CLASS TYPE:		Neural network controller
DESCRIPTION:	Objects of this class represent provide an interface to manipulate single NeuralNetwork objects.
				Adjustable learning rates and momentums are permitted.
				An arbitrary number of input nodes, hidden nodes and output nodes are permitted.
LIMITATIONS:	Only applicable to perceptrons or neural networks with one layer of hidden nodes.
				Only limited error checking is implemented. It is assumed that correct input is passed to the methods.
				Nodes uses sigmoidal activation function, so all output and expected output should be scaled when received
				   from and sent to controllers of this class so that it falls between 0.25 and 0.75.
*/

class NNController
{
	/* FIELDS ******************************************************************/


	NeuralNetwork network;		// a neural network
	int number_input_nodes;		// the number of input nodes in network
	int number_output_nodes;	// the number of output nodes in network


	/* PUBLIC METHODS **********************************************************/


	public void create_network(double learning_rate, double momentum, int num_input_nodes,
	                           int num_hidden_nodes, int num_output_nodes,
	                           double min_initial_weight, double max_initial_weight)
	// Create a new, untrained neural network
	{
		number_input_nodes = num_input_nodes;
		number_output_nodes = num_output_nodes;

		network = new NeuralNetwork(learning_rate, momentum, num_input_nodes, num_hidden_nodes,
		                            num_output_nodes, min_initial_weight, max_initial_weight);
	}

	public void load_network(double learning_rate, double momentum, int num_input_nodes,
	                         int num_hidden_nodes, int num_output_nodes, double[] weights)
	// Create a pre-trained neuralnetwork. The weights array contains the weights of
	// the hidden nodes (if any) followed by those of the output nodes.
	{
		number_input_nodes = num_input_nodes;
		number_output_nodes = num_output_nodes;
		network = new NeuralNetwork(learning_rate, momentum, num_input_nodes, num_hidden_nodes,
		                            num_output_nodes, weights);
	}

	public double[] train_network(int iterations, double[][] training_input, double[][] expected_output)
	// Train the network using the given array of possible sets of inputs and the given array
	// of corresponding sets of desired outputs. Returns an array holding the sum-of-squared-error
	// for each set of inputs averaged over the range of input sets for each iteration.
	// Output every 250 iterations to standard out
	{
		double error; // the sum-of-squared-error for each input set
		double error_sum; // the total sum-of-squared-error for each training iteration
		double[] average_error; // the average sum-of-squared-error for each training iteration
		
		average_error = new double[iterations];

		for (int i = 0; i < iterations; i++)
		{
			error_sum = 0;
			for (int j = 0; j < training_input.length; j++)
			{
				error = network.training_iteration(training_input[j], expected_output[j]);
				error_sum = error_sum + error;
			}
			average_error[i] = error_sum / training_input.length;

			// Print current status to standard out every 1000th iteration
			if (i % 1000 == 0)
			{
				int percent_done = (int) (100.0 * ((double) i) / ((double) iterations));
				System.out.println("   Epoch: " + i + " of " + iterations + " (" + percent_done +
					"%): " +  ((float) average_error[i]));
			}
		}

		return average_error;
	}

	public double get_error(double[][] given_input, double[][] expected_output)
	// Returns the error for the given set of data
	{
		double error; // the sum-of-squared-error for each input set
		double error_sum = 0; // the total sum-of-squared-error for each training iteration

		for (int j = 0; j < given_input.length; j++)
		{
			error = network.get_error(given_input[j], expected_output[j]);
			error_sum = error_sum + error;
		}

		return error_sum / given_input.length;
	}

	public double[] get_output(double[] input_values)
	// Return the output of the network for the given set of inputs
	{
		return network.get_output(input_values);
	}

	public String expose_network()
	// Returns a string containing the current inputs, weights and values of the network nodes.
	// Used during debugging
	{
		return network.print_network();
	}

	public String get_network_data()
	// Returns a string contating a vertical list of learning rate, momentum, number of
	// input nodes, number of hidden nodes, number of output nodes and all weights (hidden
	// nodes followed by output nodes).
	{
		return network.get_weight_string();
	}
}