/*
NetworksFileReader
by Cory McKay

CLASS NAME:		NetworksFileReader
CLASS TYPE:		File reader and processor
DESCRIPTION:	Loads in saved networks and parses their settings and weights into the network_data matrix.
				Uses the following format for this:
					<learning_rate>
					<momentum>
					<number_of_input_nodes>
					<number_of_hidden_nodes>
					<number_of_output_nodes>
					<hidden_node_weight_#1> NOTE: this is omitted if there is no hidden layer
					...
					<hidden_node_weight_#NHN> NOTE: this is omitted if there is no hidden layer
					<output_node_weight_#1>
					...
					<output_node_weight_#NON>
*/

import java.io.*;
import java.util.StringTokenizer;

class NetworksFileReader
{
	/* FIELDS ******************************************************************/


	private double[] network_data;		// settings and weights of a saved network
	private char[] chars;				// an array that holds all of the characters that are read in through
										//     the file
	private int next_index;				// the current location in the chars array
	private int number_of_lines;		// the number of lines in the file


	/* PUBLIC METHODS **********************************************************/


	public Object[] get_overall_network_data(String filename)
		throws IOException
	// Given the name with a file containing the network type, type of input,
	// number of bins/partials, number of networks in an array of networks, the
	// number of instruments, the min pitch and the max pitch, return this data in an array
	{
		Object[] return_data = new Object[3];

		// Initialize fields
		chars = null;
		next_index = 0;

		// Load the file into memory
		access_file(filename);

		// Read in network type and input type
		return_data[0] = get_next_string();
		return_data[1] = get_next_string();

		// Set up the data array
		int[] settings = new int[5];
		for (int i = 0; i < 5; i++)
			settings[i] = Integer.parseInt(get_next_string());
		return_data[2] = settings;

		// Return the relevant information
		return return_data;
	}



	public void read_network_file(String filename)
		throws IOException
	// Read in and parse a file containing the data specifying a neural network.
	// The format of the file is as follows:
	// <learning_rate>
	// <momentum>
	// <number_of_input_nodes>
	// <number_of_hidden_nodes>
	// <number_of_output_nodes>
	// <hidden_node_weight_#1> NOTE: this is omitted if there is no hidden layer
	// ...
	// <hidden_node_weight_#NHN> NOTE: this is omitted if there is no hidden layer
	// <output_node_weight_#1>
	// ...
	// <output_node_weight_#NON>
	{
		// Initialize fields
		chars = null;
		next_index = 0;

		// Load the file into memory
		access_file(filename);

		// Parse the contents of the file into the input_data and output_data arrays
		parse_weight_data();
	}

	public double[] get_network_data()
	// Returns the configuration and weights of a saved neural network. The first entry
	// is the learning rate, followed by the momentum, number of input nodes, number
	// of hidden nodes, number of output nodes, hidden node 1 weight (if there is a
	// hidden layer), ... , last hidden noe weight, ... , output node 1 weight, last
	// output node.
	{
		return network_data;
	}


	/* PRIVATE METHODS *********************************************************/

	
	private void access_file(String filename) 
		throws IOException
	// Instantiates chars and reads the content of the specified file into chars
	{
		File src = new File(filename);
		chars = new char[(int) src.length()];
		src = null;
		
		InputStream in;
		in = new FileInputStream(filename);
		int ch;
		for (int i = 0; (ch = in.read()) != -1; i++)
			chars[i] = (char)ch;
		in.close();
	}

	private String get_next_string()
	// Reads in one line from chars and returns it as a string. Also moves next_index to
	// the next_location in the file.
	{
		int i = next_index;
		int done = -1;

		while (done != 1)
		{
			if (i < chars.length && done != 1)
			{
				if ( chars[i] != '\n' )
					i++;
				else done = 1;
			}
			else done = 1;
		}
		String temp = new String(chars, next_index, i-next_index);
		next_index = i + 1;
		return temp;
	}

	private void find_number_lines()
	// Counts the number of lines in chars
	{
		int index_of_last_return = -1;
		number_of_lines = 0;
		for (int i = 0; i < chars.length; i++)
		{
			if (chars[i] == '\n')
			{
				number_of_lines++;
				index_of_last_return = i;
			}
		}

		// Deal with case where there is no final return in data file
		if ((chars.length - 1) != index_of_last_return)
			number_of_lines++;
	}

	private void parse_weight_data()
	// Takes the training data stored in chars and parses it so that it is contained in
	// the network_data array.
	{
		String current_line = null; // the current line of information being read
		int line_number = 0; // the number of the current line being processed

		find_number_lines();
		network_data = new double[number_of_lines];

		// Read and insert data into loaded_network	
		int i = 0; // current index of network_data
		while ( next_index < chars.length )
		{
			current_line = get_next_string();
			network_data[i] = new Double(current_line).doubleValue();
			i++;
			line_number++;
		}
	}
}