/*
NewNetworkArrayDialog
by Cory McKay

CLASS NAME:		NewNetworkArrayDialog
CLASS TYPE:		GUI - dialog box
DESCRIPTION:	This dialog box is used to get the basic settings for a new array of networks
				from the user. The user is able to set the number of networks in the array,
				the number of input, hidden and output nodes, the learning rate, the momentum,
				the minimum initial weight and the maximum initial weight for the networks.
				All networks are given the same settings unless specified otherwise with the
				ExceptionSettingsDialog dialog box.
				The was_updated method can be used to see if the OK button was pressed (as
				opposed to the Cancel button or the close box). The get_new_network_array method
				can be called to get an NNArray object that contains the properly configured
				networks with randomly generated weights.
*/


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


class NewNetworkArrayDialog
	extends JDialog
{
	/* FIELDS ******************************************************************/


	// Set to true after the OK button is pressed, is false otherwise
	private boolean update; 

	// Indicates settings of any networks that have been separately specified
	NetworkSettings[] exceptions;

	// Fields holding network configuration data
	public JTextField num_networks_field;
	public JTextField num_inputs_field;
	public JTextField num_hiddens_field;
	public JTextField num_outputs_field;
	public JTextField learning_rate_field;
	public JTextField momentum_factor_field;
	public JTextField min_init_weight_field;
	public JTextField max_init_weight_field;
	public JTextField exception_index_field;


	/* CONSTRUCTOR *************************************************************/


	public NewNetworkArrayDialog(JFrame owner)
	{
		// Give the dialog box its owner, its title and make it modal
		super(owner, "Create New Array of Networks", true);

		// Initialize the dialog box as not having been updated (to distinguish between ok and cancel)
		update = false;

		// Initialize so that all networks will have same settings
		exceptions = null;

		// Prepare panels to display dialog box contents
		Container contentPane = getContentPane();
		JPanel outer_panel = new JPanel();
		outer_panel.setLayout(new BorderLayout());
		JPanel text_panel = new JPanel();
		text_panel.setLayout(new GridLayout(8,2));
		JPanel button_panel = new JPanel();
		JPanel exception_panel = new JPanel();

		// Configure an OK button to close the dialog box
		// Sets update to true so that the calling class can check to see if OK was pressed
		JButton ok = new JButton("OK");
		ok.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				update = true;
				setVisible(false);
			}
		});

		// Configure a cancel button to close the dialog
		JButton cancel = new JButton("Cancel");
		cancel.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				setVisible(false);
			}
		});

		// Configure an exception button to specify settings for individual networks
		JButton exception_button = new JButton("Define Exception:");
		exception_button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				ExceptionSettingsDialog exception_dialog =
					new ExceptionSettingsDialog(NewNetworkArrayDialog.this, Integer.parseInt(exception_index_field.getText()));
				exception_dialog.show();

				if (exception_dialog.was_updated() == true) // take action if the OK button was pressed
				{
					NetworkSettings[] temp_exceptions
						= new NetworkSettings[Integer.parseInt(num_networks_field.getText())];

					if (exceptions == null)
					{
						for (int i = 0; i < temp_exceptions.length; i++)
						{
							if (i == Integer.parseInt(exception_index_field.getText()))
								temp_exceptions[i] = exception_dialog.get_settings();
							else temp_exceptions[i] = null;
						}
					}
					else
					{
						for (int i = 0; i < temp_exceptions.length; i++)
						{
							if (i == Integer.parseInt(exception_index_field.getText()))
								temp_exceptions[i] = exception_dialog.get_settings();
							else if (i >= exceptions.length)
								temp_exceptions[i] = null;
							else
								temp_exceptions[i] = exceptions[i];
						}
					}
					
					exceptions = temp_exceptions;
				}
			}
		});

		// Configure a button to clear exceptions
		JButton clear_exceptions_button = new JButton("Clear Exceptions");
		clear_exceptions_button.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent evt)
			{
				exceptions = null;
			}
		});

		// Initialize text fields
		num_networks_field = new JTextField("1", 3);
		num_inputs_field = new JTextField("3", 3);
		num_hiddens_field = new JTextField("4", 3);
		num_outputs_field = new JTextField("1", 3);
		learning_rate_field = new JTextField("0.1", 4);
		momentum_factor_field = new JTextField("0.1", 4);
		min_init_weight_field = new JTextField("0.3", 4);
		max_init_weight_field = new JTextField("0.7", 4);
		exception_index_field = new JTextField("0", 3);

		// Prevent user from editing the number of networks or number of input or output nodes directly
		num_networks_field.setEditable(false);
		num_inputs_field.setEditable(false);
		num_outputs_field.setEditable(false);
		
		// Set up text_panel
		text_panel.add(new JLabel("Num of Networks:"));
		text_panel.add(num_networks_field);
		text_panel.add(new JLabel("Inputs Nodes:"));
		text_panel.add(num_inputs_field);
		text_panel.add(new JLabel("Hidden Nodes:"));
		text_panel.add(num_hiddens_field);
		text_panel.add(new JLabel("Outputs Noes:"));
		text_panel.add(num_outputs_field);
		text_panel.add(new JLabel("Learning Rate:"));
		text_panel.add(learning_rate_field);
		text_panel.add(new JLabel("Momentum:"));
		text_panel.add(momentum_factor_field);
		text_panel.add(new JLabel("Min Init Weight:"));
		text_panel.add(min_init_weight_field);
		text_panel.add(new JLabel("Max Init Weight:"));
		text_panel.add(max_init_weight_field);

		// Set up exception_panel
		exception_panel.add(exception_button);
		exception_panel.add(exception_index_field);
		exception_panel.add(clear_exceptions_button);

		// Set up button_panel
		button_panel.add(ok);
		button_panel.add(cancel);

		// Display dialog box contents
		outer_panel.add(text_panel, BorderLayout.NORTH);
		outer_panel.add(exception_panel, BorderLayout.CENTER);
		outer_panel.add(button_panel, BorderLayout.SOUTH);
		contentPane.add(outer_panel);
		setSize(340, 270);
	}


	/* PUBLIC METHODS **********************************************************/


	public boolean was_updated()
	// Returns true if the OK button was pressed the last time the dialog box was visible.
	// Returns false otherwise. Resets the update field to false.
	{
		boolean temp = update;
		update = false;
		return temp;
	}


	/***************************************************************************/


	public NNArray generate_new_network_array()
	// Returns an NNArray created based on the specifications given in this dialog box.
	// Weights are randomly generated.
	{
		NetworkSettings[] settings = new NetworkSettings[Integer.parseInt(num_networks_field.getText())];

		if (exceptions != null)
		{
			for (int i = 0; i < exceptions.length; i++)
			{
				if (exceptions[i] == null)
				{
					settings[i] = new NetworkSettings
					(
						Double.parseDouble(learning_rate_field.getText()),
						Double.parseDouble(momentum_factor_field.getText()),
						Integer.parseInt(num_inputs_field.getText()),
						Integer.parseInt(num_hiddens_field.getText()),
						Integer.parseInt(num_outputs_field.getText()),
						Double.parseDouble(min_init_weight_field.getText()),
						Double.parseDouble(max_init_weight_field.getText())
					);
				}
				else settings[i] = exceptions[i];
			}
		}

		else
		{
			for (int i = 0; i < Integer.parseInt(num_networks_field.getText()); i++)
			{
				settings[i] = new NetworkSettings
				(
					Double.parseDouble(learning_rate_field.getText()),
					Double.parseDouble(momentum_factor_field.getText()),
					Integer.parseInt(num_inputs_field.getText()),
					Integer.parseInt(num_hiddens_field.getText()),
					Integer.parseInt(num_outputs_field.getText()),
					Double.parseDouble(min_init_weight_field.getText()),
					Double.parseDouble(max_init_weight_field.getText())
				);
			}
		}

		return new NNArray(settings);
	}
}